@@ -23,8 +23,6 @@ import inspect
 import collections
 import traceback
 
-from PyQt5.QtWebKit import QWebSettings
-
 from qutebrowser.commands import cmdexc, argparser
 from qutebrowser.utils import (log, utils, message, docutils, objreg,
                                usertypes, typing)
@@ -35,8 +33,8 @@ class ArgInfo:
 
     """Information about an argument."""
 
-    def __init__(self, win_id=False, count=False, flag=None, hide=False,
-                 metavar=None, completion=None, choices=None):
+    def __init__(self, win_id=False, count=False, hide=False, metavar=None,
+                 flag=None, completion=None, choices=None):
         if win_id and count:
             raise TypeError("Argument marked as both count/win_id!")
         self.win_id = win_id
@@ -75,27 +73,25 @@ class Command:
         deprecated: False, or a string to describe why a command is deprecated.
         desc: The description of the command.
         handler: The handler function to call.
-        completion: Completions to use for arguments, as a list of strings.
         debug: Whether this is a debugging command (only shown with --debug).
         parser: The ArgumentParser to use to parse this command.
         flags_with_args: A list of flags which take an argument.
         no_cmd_split: If true, ';;' to split sub-commands is ignored.
         backend: Which backend the command works with (or None if it works with
                  both)
+        no_replace_variables: Don't replace variables like {url}
         _qute_args: The saved data from @cmdutils.argument
-        _needs_js: Whether the command needs javascript enabled
         _modes: The modes the command can be executed in.
-        _not_modes: The modes the command can not be executed in.
         _count: The count set for the command.
         _instance: The object to bind 'self' to.
         _scope: The scope to get _instance for in the object registry.
     """
 
     def __init__(self, *, handler, name, instance=None, maxsplit=None,
-                 hide=False, modes=None, not_modes=None, needs_js=False,
-                 debug=False, ignore_args=False, deprecated=False,
-                 no_cmd_split=False, star_args_optional=False, scope='global',
-                 backend=None):
+                 hide=False, modes=None, not_modes=None, debug=False,
+                 ignore_args=False, deprecated=False, no_cmd_split=False,
+                 star_args_optional=False, scope='global', backend=None,
+                 no_replace_variables=False):
         # I really don't know how to solve this in a better way, I tried.
         # pylint: disable=too-many-locals
         if modes is not None and not_modes is not None:
@@ -104,10 +100,14 @@ class Command:
             for m in modes:
                 if not isinstance(m, usertypes.KeyMode):
                     raise TypeError("Mode {} is no KeyMode member!".format(m))
-        if not_modes is not None:
+            self._modes = set(modes)
+        elif not_modes is not None:
             for m in not_modes:
                 if not isinstance(m, usertypes.KeyMode):
                     raise TypeError("Mode {} is no KeyMode member!".format(m))
+            self._modes = set(usertypes.KeyMode).difference(not_modes)
+        else:
+            self._modes = set(usertypes.KeyMode)
         if scope != 'global' and instance is None:
             raise ValueError("Setting scope without setting instance makes "
                              "no sense!")
@@ -117,16 +117,14 @@ class Command:
         self.hide = hide
         self.deprecated = deprecated
         self._instance = instance
-        self._modes = modes
-        self._not_modes = not_modes
         self._scope = scope
-        self._needs_js = needs_js
         self._star_args_optional = star_args_optional
         self.debug = debug
         self.ignore_args = ignore_args
         self.handler = handler
         self.no_cmd_split = no_cmd_split
         self.backend = backend
+        self.no_replace_variables = no_replace_variables
 
         self.docparser = docutils.DocstringParser(handler)
         self.parser = argparser.ArgumentParser(
@@ -148,13 +146,7 @@ class Command:
         self._qute_args = getattr(self.handler, 'qute_args', {})
         self.handler.qute_args = None
 
-        args = self._inspect_func()
-
-        self.completion = []
-        for arg in args:
-            arg_completion = self.get_arg_info(arg).completion
-            if arg_completion is not None:
-                self.completion.append(arg_completion)
+        self._inspect_func()
 
     def _check_prerequisites(self, win_id):
         """Check if the command is permitted to run currently.
@@ -164,36 +156,17 @@ class Command:
         """
         mode_manager = objreg.get('mode-manager', scope='window',
                                   window=win_id)
-        curmode = mode_manager.mode
-        if self._modes is not None and curmode not in self._modes:
-            mode_names = '/'.join(mode.name for mode in self._modes)
-            raise cmdexc.PrerequisitesError(
-                "{}: This command is only allowed in {} mode.".format(
-                    self.name, mode_names))
-        elif self._not_modes is not None and curmode in self._not_modes:
-            mode_names = '/'.join(mode.name for mode in self._not_modes)
-            raise cmdexc.PrerequisitesError(
-                "{}: This command is not allowed in {} mode.".format(
-                    self.name, mode_names))
+        self.validate_mode(mode_manager.mode)
 
-        if self._needs_js and not QWebSettings.globalSettings().testAttribute(
-                QWebSettings.JavascriptEnabled):
-            raise cmdexc.PrerequisitesError(
-                "{}: This command needs javascript enabled.".format(self.name))
-
-        backend_mapping = {
-            'webkit': usertypes.Backend.QtWebKit,
-            'webengine': usertypes.Backend.QtWebEngine,
-        }
-        used_backend = backend_mapping[objreg.get('args').backend]
+        used_backend = usertypes.arg2backend[objreg.get('args').backend]
         if self.backend is not None and used_backend != self.backend:
             raise cmdexc.PrerequisitesError(
                 "{}: Only available with {} "
                 "backend.".format(self.name, self.backend.name))
 
         if self.deprecated:
-            message.warning(win_id, '{} is deprecated - {}'.format(
-                self.name, self.deprecated))
+            message.warning('{} is deprecated - {}'.format(self.name,
+                                                           self.deprecated))
 
     def _check_func(self):
         """Make sure the function parameters don't violate any rules."""
@@ -213,6 +186,11 @@ class Command:
         """Get an ArgInfo tuple for the given inspect.Parameter."""
         return self._qute_args.get(param.name, ArgInfo())
 
+    def get_pos_arg_info(self, pos):
+        """Get an ArgInfo tuple for the given positional parameter."""
+        name = self.pos_args[pos][0]
+        return self._qute_args.get(name, ArgInfo())
+
     def _inspect_special_param(self, param):
         """Check if the given parameter is a special one.
 
@@ -249,10 +227,21 @@ class Command:
 
         if not self.ignore_args:
             for param in signature.parameters.values():
+                # https://docs.python.org/3/library/inspect.html#inspect.Parameter.kind
+                # "Python has no explicit syntax for defining positional-only
+                # parameters, but many built-in and extension module functions
+                # (especially those that accept only one or two parameters)
+                # accept them."
+                assert param.kind != inspect.Parameter.POSITIONAL_ONLY
                 if param.name == 'self':
                     continue
                 if self._inspect_special_param(param):
                     continue
+                if (param.kind == inspect.Parameter.KEYWORD_ONLY and
+                        param.default is inspect.Parameter.empty):
+                    raise TypeError("{}: handler has keyword only argument "
+                                    "{!r} without default!".format(self.name,
+                                                                   param.name))
                 typ = self._get_type(param)
                 is_bool = typ is bool
                 kwargs = self._param_to_argparse_kwargs(param, is_bool)
@@ -345,12 +334,17 @@ class Command:
         Args:
             param: The inspect.Parameter to look at.
         """
+        arginfo = self.get_arg_info(param)
         if param.annotation is not inspect.Parameter.empty:
             return param.annotation
-        elif param.default is None or param.default is inspect.Parameter.empty:
+        elif param.default not in [None, inspect.Parameter.empty]:
+            return type(param.default)
+        elif arginfo.count or arginfo.win_id or param.kind in [
+                inspect.Parameter.VAR_POSITIONAL,
+                inspect.Parameter.VAR_KEYWORD]:
             return None
         else:
-            return type(param.default)
+            return str
 
     def _get_self_arg(self, win_id, param, args):
         """Get the self argument for a function call.
@@ -437,10 +431,10 @@ class Command:
         elif typ is str:
             choices = self.get_arg_info(param).choices
             value = argparser.type_conv(param, typ, value, str_choices=choices)
-        elif typ is None:
-            pass
         elif typ is bool:  # no type conversion for flags
             assert isinstance(value, bool)
+        elif typ is None:
+            pass
         else:
             value = argparser.type_conv(param, typ, value)
 
@@ -516,7 +510,7 @@ class Command:
         try:
             self.namespace = self.parser.parse_args(args)
         except argparser.ArgumentParserError as e:
-            message.error(win_id, '{}: {}'.format(self.name, e),
+            message.error('{}: {}'.format(self.name, e),
                           stack=traceback.format_exc())
             return
         except argparser.ArgumentParserExit as e:
@@ -529,3 +523,15 @@ class Command:
         log.commands.debug('Calling {}'.format(
             debug_utils.format_call(self.handler, posargs, kwargs)))
         self.handler(*posargs, **kwargs)
+
+    def validate_mode(self, mode):
+        """Raise cmdexc.PrerequisitesError unless allowed in the given mode.
+
+        Args:
+            mode: The usertypes.KeyMode to check.
+        """
+        if mode not in self._modes:
+            mode_names = '/'.join(sorted(m.name for m in self._modes))
+            raise cmdexc.PrerequisitesError(
+                "{}: This command is only allowed in {} mode, not {}.".format(
+                    self.name, mode_names, mode.name))
