@@ -29,7 +29,7 @@ from qutebrowser.utils import message, log, objreg, standarddir
 from qutebrowser.commands import runners
 from qutebrowser.config import config
 from qutebrowser.misc import guiprocess
-from qutebrowser.browser.webkit import downloads
+from qutebrowser.browser import downloads
 
 
 class _QtFIFOReader(QObject):
@@ -65,9 +65,13 @@ class _QtFIFOReader(QObject):
         """(Try to) read a line from the FIFO."""
         log.procs.debug("QSocketNotifier triggered!")
         self._notifier.setEnabled(False)
-        for line in self._fifo:
-            self.got_line.emit(line.rstrip('\r\n'))
-        self._notifier.setEnabled(True)
+        try:
+            for line in self._fifo:
+                self.got_line.emit(line.rstrip('\r\n'))
+                self._notifier.setEnabled(True)
+        except UnicodeDecodeError as e:
+            log.misc.error("Invalid unicode in userscript output: {}"
+                           .format(e))
 
     def cleanup(self):
         """Clean up so the FIFO can be closed."""
@@ -84,7 +88,6 @@ class _BaseUserscriptRunner(QObject):
     Attributes:
         _filepath: The path of the file/FIFO which is being read.
         _proc: The GUIProcess which is being executed.
-        _win_id: The window ID this runner is associated with.
         _cleaned_up: Whether temporary files were cleaned up.
         _text_stored: Set when the page text was stored async.
         _html_stored: Set when the page html was stored async.
@@ -99,10 +102,9 @@ class _BaseUserscriptRunner(QObject):
     got_cmd = pyqtSignal(str)
     finished = pyqtSignal()
 
-    def __init__(self, win_id, parent=None):
+    def __init__(self, parent=None):
         super().__init__(parent)
         self._cleaned_up = False
-        self._win_id = win_id
         self._filepath = None
         self._proc = None
         self._env = {}
@@ -151,7 +153,7 @@ class _BaseUserscriptRunner(QObject):
         self._env['QUTE_FIFO'] = self._filepath
         if env is not None:
             self._env.update(env)
-        self._proc = guiprocess.GUIProcess(self._win_id, 'userscript',
+        self._proc = guiprocess.GUIProcess('userscript',
                                            additional_env=self._env,
                                            verbose=verbose, parent=self)
         self._proc.finished.connect(self.on_proc_finished)
@@ -175,9 +177,8 @@ class _BaseUserscriptRunner(QObject):
             except OSError as e:
                 # NOTE: Do not replace this with "raise CommandError" as it's
                 # executed async.
-                message.error(
-                    self._win_id, "Failed to delete tempfile {} ({})!".format(
-                        fn, e))
+                message.error("Failed to delete tempfile {} ({})!".format(
+                    fn, e))
         self._filepath = None
         self._proc = None
         self._env = {}
@@ -223,8 +224,8 @@ class _POSIXUserscriptRunner(_BaseUserscriptRunner):
         _reader: The _QtFIFOReader instance.
     """
 
-    def __init__(self, win_id, parent=None):
-        super().__init__(win_id, parent)
+    def __init__(self, parent=None):
+        super().__init__(parent)
         self._reader = None
 
     def prepare_run(self, *args, **kwargs):
@@ -242,8 +243,7 @@ class _POSIXUserscriptRunner(_BaseUserscriptRunner):
             # pylint: disable=no-member,useless-suppression
             os.mkfifo(self._filepath)
         except OSError as e:
-            message.error(self._win_id,
-                          "Error while creating FIFO: {}".format(e))
+            message.error("Error while creating FIFO: {}".format(e))
             return
 
         self._reader = _QtFIFOReader(self._filepath)
@@ -280,15 +280,8 @@ class _WindowsUserscriptRunner(_BaseUserscriptRunner):
 
     This also means the userscript *has* to use >> (append) rather than >
     (overwrite) to write to the file!
-
-    Attributes:
-        _oshandle: The oshandle of the temp file.
     """
 
-    def __init__(self, win_id, parent=None):
-        super().__init__(win_id, parent)
-        self._oshandle = None
-
     def _cleanup(self):
         """Clean up temporary files after the userscript finished."""
         if self._cleaned_up:
@@ -300,13 +293,11 @@ class _WindowsUserscriptRunner(_BaseUserscriptRunner):
                     self.got_cmd.emit(line.rstrip())
         except OSError:
             log.procs.exception("Failed to read command file!")
+        except UnicodeDecodeError as e:
+            log.misc.error("Invalid unicode in userscript output: {}"
+                           .format(e))
 
-        try:
-            os.close(self._oshandle)
-        except OSError:
-            log.procs.exception("Failed to close file handle!")
         super()._cleanup()
-        self._oshandle = None
         self.finished.emit()
 
     @pyqtSlot()
@@ -323,14 +314,42 @@ class _WindowsUserscriptRunner(_BaseUserscriptRunner):
         self._kwargs = kwargs
 
         try:
-            self._oshandle, self._filepath = tempfile.mkstemp(text=True)
+            handle = tempfile.NamedTemporaryFile(delete=False)
+            handle.close()
+            self._filepath = handle.name
         except OSError as e:
-            message.error(self._win_id, "Error while creating tempfile: "
-                                        "{}".format(e))
+            message.error("Error while creating tempfile: {}".format(e))
             return
 
 
-class UnsupportedError(Exception):
+class Error(Exception):
+
+    """Base class for userscript exceptions."""
+
+
+class NotFoundError(Error):
+
+    """Raised when spawning a userscript that doesn't exist.
+
+    Attributes:
+        script_name: name of the userscript as called
+        paths: path names that were searched for the userscript
+    """
+
+    def __init__(self, script_name, paths=None):
+        super().__init__()
+        self.script_name = script_name
+        self.paths = paths
+
+    def __str__(self):
+        msg = "Userscript '{}' not found".format(self.script_name)
+        if self.paths:
+            msg += " in userscript directories {}".format(
+                ', '.join(repr(path) for path in self.paths))
+        return msg
+
+
+class UnsupportedError(Error):
 
     """Raised when userscripts aren't supported on this platform."""
 
@@ -338,12 +357,37 @@ class UnsupportedError(Exception):
         return "Userscripts are not supported on this platform!"
 
 
+def _lookup_path(cmd):
+    """Search userscript directories for given command.
+
+    Raises:
+        NotFoundError if the command could not be found.
+
+    Args:
+        cmd: The command to look for.
+
+    Returns:
+        A path to the userscript.
+    """
+    directories = [
+        os.path.join(standarddir.data(), "userscripts"),
+        os.path.join(standarddir.system_data(), "userscripts"),
+    ]
+    for directory in directories:
+        cmd_path = os.path.join(directory, cmd)
+        if os.path.exists(cmd_path):
+            return cmd_path
+
+    raise NotFoundError(cmd, directories)
+
+
 def run_async(tab, cmd, *args, win_id, env, verbose=False):
     """Run a userscript after dumping page html/source.
 
     Raises:
         UnsupportedError if userscripts are not supported on the current
         platform.
+        NotFoundError if the command could not be found.
 
     Args:
         tab: The WebKitTab/WebEngineTab to get the source from.
@@ -358,9 +402,9 @@ def run_async(tab, cmd, *args, win_id, env, verbose=False):
     commandrunner = runners.CommandRunner(win_id, parent=tabbed_browser)
 
     if os.name == 'posix':
-        runner = _POSIXUserscriptRunner(win_id, tabbed_browser)
+        runner = _POSIXUserscriptRunner(tabbed_browser)
     elif os.name == 'nt':  # pragma: no cover
-        runner = _WindowsUserscriptRunner(win_id, tabbed_browser)
+        runner = _WindowsUserscriptRunner(tabbed_browser)
     else:  # pragma: no cover
         raise UnsupportedError
 
@@ -371,25 +415,20 @@ def run_async(tab, cmd, *args, win_id, env, verbose=False):
     user_agent = config.get('network', 'user-agent')
     if user_agent is not None:
         env['QUTE_USER_AGENT'] = user_agent
-    config_dir = standarddir.config()
-    if config_dir is not None:
-        env['QUTE_CONFIG_DIR'] = config_dir
-    data_dir = standarddir.data()
-    if data_dir is not None:
-        env['QUTE_DATA_DIR'] = data_dir
-    download_dir = downloads.download_dir()
-    if download_dir is not None:
-        env['QUTE_DOWNLOAD_DIR'] = download_dir
+
+    env['QUTE_CONFIG_DIR'] = standarddir.config()
+    env['QUTE_DATA_DIR'] = standarddir.data()
+    env['QUTE_DOWNLOAD_DIR'] = downloads.download_dir()
+
     cmd_path = os.path.expanduser(cmd)
 
     # if cmd is not given as an absolute path, look it up
     # ~/.local/share/qutebrowser/userscripts (or $XDG_DATA_DIR)
     if not os.path.isabs(cmd_path):
         log.misc.debug("{} is no absolute path".format(cmd_path))
-        cmd_path = os.path.join(standarddir.data(), "userscripts", cmd)
-        if not os.path.exists(cmd_path):
-            cmd_path = os.path.join(standarddir.system_data(),
-                                    "userscripts", cmd)
+        cmd_path = _lookup_path(cmd)
+    elif not os.path.exists(cmd_path):
+        raise NotFoundError(cmd_path)
     log.misc.debug("Userscript to run: {}".format(cmd_path))
 
     runner.finished.connect(commandrunner.deleteLater)
