@@ -24,12 +24,12 @@ subclasses to provide completions.
 """
 
 from PyQt5.QtWidgets import QStyle, QTreeView, QSizePolicy
-from PyQt5.QtCore import pyqtSlot, pyqtSignal, Qt, QItemSelectionModel
+from PyQt5.QtCore import pyqtSlot, pyqtSignal, Qt, QItemSelectionModel, QSize
 
 from qutebrowser.config import config, style
-from qutebrowser.completion import completiondelegate, completer
+from qutebrowser.completion import completiondelegate
 from qutebrowser.completion.models import base
-from qutebrowser.utils import qtutils, objreg, utils, usertypes
+from qutebrowser.utils import utils, usertypes, objreg
 from qutebrowser.commands import cmdexc, cmdutils
 
 
@@ -41,15 +41,16 @@ class CompletionView(QTreeView):
     headers, and children show as flat list.
 
     Attributes:
-        enabled: Whether showing the CompletionView is enabled.
         _win_id: The ID of the window this CompletionView is associated with.
         _height: The height to use for the CompletionView.
         _height_perc: Either None or a percentage if height should be relative.
         _delegate: The item delegate used.
         _column_widths: A list of column widths, in percent.
+        _active: Whether a selection is active.
 
     Signals:
-        resize_completion: Emitted when the completion should be resized.
+        update_geometry: Emitted when the completion should be resized.
+        selection_changed: Emitted when the completion item selection changes.
     """
 
     # Drawing the item foreground will be done by CompletionItemDelegate, so we
@@ -101,28 +102,23 @@ class CompletionView(QTreeView):
         }
     """
 
-    resize_completion = pyqtSignal()
+    update_geometry = pyqtSignal()
+    selection_changed = pyqtSignal(str)
 
     def __init__(self, win_id, parent=None):
         super().__init__(parent)
         self._win_id = win_id
-        objreg.register('completion', self, scope='window', window=win_id)
-        cmd = objreg.get('status-command', scope='window', window=win_id)
-        completer_obj = completer.Completer(cmd, win_id, self)
-        completer_obj.next_prev_item.connect(self.on_next_prev_item)
-        objreg.register('completer', completer_obj, scope='window',
-                        window=win_id)
-        self.enabled = config.get('completion', 'show')
-        objreg.get('config').changed.connect(self.set_enabled)
         # FIXME handle new aliases.
         # objreg.get('config').changed.connect(self.init_command_completion)
+        objreg.get('config').changed.connect(self._on_config_changed)
 
         self._column_widths = base.BaseCompletionModel.COLUMN_WIDTHS
+        self._active = False
 
         self._delegate = completiondelegate.CompletionItemDelegate(self)
         self.setItemDelegate(self._delegate)
         style.set_register_stylesheet(self)
-        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
+        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
         self.setHeaderHidden(True)
         self.setAlternatingRowColors(True)
         self.setIndentation(0)
@@ -143,6 +139,13 @@ class CompletionView(QTreeView):
     def __repr__(self):
         return utils.get_repr(self)
 
+    @pyqtSlot(str, str)
+    def _on_config_changed(self, section, option):
+        if section != 'completion':
+            return
+        if option in ['height', 'shrink']:
+            self.update_geometry.emit()
+
     def _resize_columns(self):
         """Resize the completion columns based on column_widths."""
         width = self.size().width()
@@ -184,87 +187,161 @@ class CompletionView(QTreeView):
                 # Item is a real item, not a category header -> success
                 return idx
 
-    @pyqtSlot(bool)
-    def on_next_prev_item(self, prev):
-        """Handle a tab press for the CompletionView.
-
-        Select the previous/next item and write the new text to the
-        statusbar.
-
-        Called from the Completer's next_prev_item signal.
+    def _next_category_idx(self, upwards):
+        """Get the index of the previous/next category.
 
         Args:
-            prev: True for prev item, False for next one.
+            upwards: Get previous item, not next.
+
+        Return:
+            A QModelIndex.
         """
-        if not self.isVisible():
-            # No completion running at the moment, ignore keypress
+        idx = self.selectionModel().currentIndex()
+        if not idx.isValid():
+            return self._next_idx(upwards).sibling(0, 0)
+        idx = idx.parent()
+        direction = -1 if upwards else 1
+        while True:
+            idx = idx.sibling(idx.row() + direction, 0)
+            if not idx.isValid() and upwards:
+                # wrap around to the first item of the last category
+                return self.model().last_item().sibling(0, 0)
+            elif not idx.isValid() and not upwards:
+                # wrap around to the first item of the first category
+                idx = self.model().first_item()
+                self.scrollTo(idx.parent())
+                return idx
+            elif idx.isValid() and idx.child(0, 0).isValid():
+                # scroll to ensure the category is visible
+                self.scrollTo(idx)
+                return idx.child(0, 0)
+
+    @cmdutils.register(instance='completion', hide=True,
+                       modes=[usertypes.KeyMode.command], scope='window')
+    @cmdutils.argument('which', choices=['next', 'prev', 'next-category',
+                                         'prev-category'])
+    def completion_item_focus(self, which):
+        """Shift the focus of the completion menu to another item.
+
+        Args:
+            which: 'next', 'prev', 'next-category', or 'prev-category'.
+        """
+        if not self._active:
             return
-        idx = self._next_idx(prev)
-        qtutils.ensure_valid(idx)
-        self.selectionModel().setCurrentIndex(
+        selmodel = self.selectionModel()
+
+        if which == 'next':
+            idx = self._next_idx(upwards=False)
+        elif which == 'prev':
+            idx = self._next_idx(upwards=True)
+        elif which == 'next-category':
+            idx = self._next_category_idx(upwards=False)
+        elif which == 'prev-category':
+            idx = self._next_category_idx(upwards=True)
+        else:  # pragma: no cover
+            raise ValueError("Invalid 'which' value {!r}".format(which))
+
+        if not idx.isValid():
+            return
+
+        selmodel.setCurrentIndex(
             idx, QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows)
 
-    def set_model(self, model):
+        count = self.model().count()
+        if count == 0:
+            self.hide()
+        elif count == 1 and config.get('completion', 'quick-complete'):
+            self.hide()
+        elif config.get('completion', 'show') == 'auto':
+            self.show()
+
+    def set_model(self, model, pattern=None):
         """Switch completion to a new model.
 
         Called from on_update_completion().
 
         Args:
             model: The model to use.
+            pattern: The filter pattern to set (what the user entered).
         """
+        if model is None:
+            self._active = False
+            self.hide()
+            return
+
         old_model = self.model()
-        sel_model = self.selectionModel()
+        if model is not old_model:
+            sel_model = self.selectionModel()
 
-        self.setModel(model)
+            self.setModel(model)
+            self._active = True
 
-        if sel_model is not None:
-            sel_model.deleteLater()
-        if old_model is not None:
-            old_model.deleteLater()
+            if sel_model is not None:
+                sel_model.deleteLater()
+            if old_model is not None:
+                old_model.deleteLater()
+
+        if (config.get('completion', 'show') == 'always' and
+                model.count() > 0):
+            self.show()
+        else:
+            self.hide()
 
         for i in range(model.rowCount()):
             self.expand(model.index(i, 0))
 
+        if pattern is not None:
+            model.set_pattern(pattern)
+
         self._column_widths = model.srcmodel.COLUMN_WIDTHS
         self._resize_columns()
-        self.maybe_resize_completion()
+        self._maybe_update_geometry()
 
-    def set_pattern(self, pattern):
-        """Set the completion pattern for the current model.
-
-        Called from on_update_completion().
-
-        Args:
-            pattern: The filter pattern to set (what the user entered).
-        """
-        self.model().set_pattern(pattern)
-        self.maybe_resize_completion()
-
-    @pyqtSlot()
-    def maybe_resize_completion(self):
-        """Emit the resize_completion signal if the config says so."""
+    def _maybe_update_geometry(self):
+        """Emit the update_geometry signal if the config says so."""
         if config.get('completion', 'shrink'):
-            self.resize_completion.emit()
-
-    @config.change_filter('completion', 'show')
-    def set_enabled(self):
-        """Update self.enabled when the config changed."""
-        self.enabled = config.get('completion', 'show')
+            self.update_geometry.emit()
 
     @pyqtSlot()
     def on_clear_completion_selection(self):
         """Clear the selection model when an item is activated."""
+        self.hide()
         selmod = self.selectionModel()
         if selmod is not None:
             selmod.clearSelection()
             selmod.clearCurrentIndex()
 
+    def sizeHint(self):
+        """Get the completion size according to the config."""
+        # Get the configured height/percentage.
+        confheight = str(config.get('completion', 'height'))
+        if confheight.endswith('%'):
+            perc = int(confheight.rstrip('%'))
+            height = self.window().height() * perc / 100
+        else:
+            height = int(confheight)
+        # Shrink to content size if needed and shrinking is enabled
+        if config.get('completion', 'shrink'):
+            contents_height = (
+                self.viewportSizeHint().height() +
+                self.horizontalScrollBar().sizeHint().height())
+            if contents_height <= height:
+                height = contents_height
+        else:
+            contents_height = -1
+        # The width isn't really relevant as we're expanding anyways.
+        return QSize(-1, height)
+
     def selectionChanged(self, selected, deselected):
         """Extend selectionChanged to call completers selection_changed."""
+        if not self._active:
+            return
         super().selectionChanged(selected, deselected)
-        completer_obj = objreg.get('completer', scope='window',
-                                   window=self._win_id)
-        completer_obj.selection_changed(selected, deselected)
+        indexes = selected.indexes()
+        if not indexes:
+            return
+        data = self.model().data(indexes[0])
+        self.selection_changed.emit(data)
 
     def resizeEvent(self, e):
         """Extend resizeEvent to adjust column size."""
@@ -273,7 +350,7 @@ class CompletionView(QTreeView):
 
     def showEvent(self, e):
         """Adjust the completion size and scroll when it's freshly shown."""
-        self.resize_completion.emit()
+        self.update_geometry.emit()
         scrollbar = self.verticalScrollBar()
         if scrollbar is not None:
             scrollbar.setValue(scrollbar.minimum())
